function [zM,uM,bV] = stap(xV, m, p, M, tau_max, K)
% [zM,uM,bV] = stap(xV, m, p, M, tau_max, K)
% 
% STAP: Statically Transformed Autoregressive Process
% STAP generates stochastic surrogates for a given time series as 
% statically transformed realisations of a Gaussian (autoregressive) 
% process. The generated surrogate data have the same amplitude 
% distribution (marginal cdf) and autocorrelation as the original time 
% series.
% INPUTS:
% -------
% xV	  : the given time series 
% m	      : order of the polynomial to approximate the sample transform
% p	      : order of the AR model to generate the Gaussian time series
% M       : number of surrogates to be generated
% tau_max : maximum lag (to be used in the comparison of autocorrelations)
% K       : number of repetitions to be made (to find the "best" AR model)	
%
% OUTPUTS
% -------
% zM	  : matrix of generated x-like surrogate data ('M' columns)
% uM	  : matrix of generated Gaussian surrogate data ('M' columns)
% bV      : vector of parameters of the "best" AR model
%
% Description of the algorithm and input parameters
% -------------------------------------------------
% First, the transform from marginal normal cdf to the marginal cdf of the 
% given time series 'xV' is estimated by a polynomial of order 'm' (the
% solution is the parameter vector 'aV'). Then the autocorrelation 'r_uV' 
% of the underlying AR-process is given as a function (defined as 
% "inverse transform") of the autocorrelation of 'xV', 'r_xV'. 
% This function is a polynomial also of order 'm' and the coefficients 'cV'
% are given analytically in terms of 'r_xV' and 'aV'. The parameters 'bV' 
% of the AR model of order 'p' are computed from 'r_uV' (using the Levinson 
% algorithm). The AR-time series 'uV' is generated and transformed to 'zV' 
% (by rank ordering), so that 'zV' has the same cdf as 'xV' and about the 
% same autocorrelation (however, 'r_zV' may vary substantially from 
% 'r_xV' at this stage). The above process is done 'K' times to obtain an 
% ensemble of candidate AR-processes. Then the most proper one is chosen in 
% the sense that the corresponding 'r_zV' matches best 'r_xV' considering 
% lags up to a given limit 'tau_max'. Using the parameters 'bV' of the "best" 
% AR-model, the 'M' realisations are generated and transformed to match the 
% cdf of 'xV'.
% 
% NOTE:
% If the sixth argument is not specified, then 'K' = 'M'.
% If also the fifth argument is not specified, then 'tau_max' = 'p'.

if nargin == 5
    K = M;
elseif nargin == 4
    K = M;
        tau_max = p;
end
ntrans = 100;
maxrepeat = 200;

n = length(xV);
[oxV,ixV] = sort(xV);
r_xV = xcorr(xV-mean(xV),tau_max,'coeff');
r_xV(1:tau_max) = [];  % the autocorrelation of xV

bM = zeros(p+1,K); % The parameters of the K candidate AR models
r_zM = zeros(tau_max+1,K); % the autocorrelation for the K candidate 
                           % transformed time series from the K AR models 
for j=1:K
	nrepeat = 1;     % Make repetitions if needed to find valid AR model
	successflag = 0; % 1 if a valid r_u is estimated
	disp(['   Sample ',int2str(j),' for STAP ...'])
    while nrepeat <= maxrepeat & successflag == 0
        wV = randn(n,1);
        [owV,iwV] = sort(wV);
        %%% Unconstrained polynomial approximation (OLS)
        pwxV = polyfit(owV,oxV,m); % Transform of margical cdf's (pol.appr.)
        pwxV = pwxV(:);
        [invr_xM,cV] = invtranscorcoef(r_xV,flipud(pwxV));
        %%% Choose the proper solution for autocorrelation
        invr_xV = zeros(p+1,1);
        invr_xV(1) = r_xV(1);
        for i=2:p+1
            inowV = find(imag(invr_xM(i,:))==0 & abs(invr_xM(i,:))<=1);
            if length(inowV) == 1
                successflag=1;
                invr_xV(i) = invr_xM(i,inowV);
            else
                successflag=0;
                nrepeat = nrepeat+1;
                if length(inowV)>1
	                disp(['	Trial No ',int2str(nrepeat),': r_x(',int2str(i-1),...
                            ') has ',int2str(length(inowV)),' solutions'])
                else   
	                disp(['	Trial No ',int2str(nrepeat),': r_x(',int2str(i-1),...
                            ') has no solutions']) 
                end
                break;
            end
        end
        %%%%% If a proper solution for r_u is found compute AR-parameters
        if successflag == 1
            [bV,E]=levinson(invr_xV,p);
            rootbV = roots(bV);
            if any(abs(rootbV) >= 1)
                bV = polystab(bV);
                disp(['		Try ',int2str(nrepeat),': Root(s) > 1, polystab'])
                if any(abs(roots(bV)) >= 1)
                    successflag = 0;
	                disp(['		Try ',int2str(nrepeat),...
                            ': Root(s) > 1, no solution'])
                end
            end
        end
    end
    if nrepeat > maxrepeat
        error(['No solution found within ',int2str(maxrepeat),' trials.'])
    end
    %%%%% Generate the 'uV' time series and trasnform it to 'zV'
    uV = zeros(n+p+ntrans,1);
    uV(1:p) = randn(p,1);
    for jj=p+1:n+p+ntrans
        uV(jj) = - bV(2:p+1) * uV(jj-1:-1:jj-p) + randn;
    end
    uV = uV(p+ntrans+1:n+p+ntrans);
    [T,iuV] = sort(uV);
    [T,juV] = sort(iuV);
    zV = oxV(juV); 
    %%%%% Compute also the autocorrelation of each 'zV' for later comparisons
    r_zV = xcorr(zV-mean(zV),tau_max,'coeff');
    r_zV(1:tau_max) = [];
    r_zM(:,j) = r_zV;
    bM(:,j)=bV';
end
%%%%% Find the "best" AR model that gives 'zV' with 'r_z' closest to 'r_x'
jbestV = zeros(tau_max,1);
for k=1:tau_max
    taufitV = [1:k]';
    dr_zM = r_xV(taufitV+1) * ones(1,K) - r_zM(taufitV+1,:);
    dautV = zeros(K,1);
    for j=1:K
        dautV(j) = norm(dr_zM(:,j));
    end
    [dbest,jbestV(k)]=min(dautV);     
end
%=== 'jbestV' contains the indices of "best" candidates for each max lag
%=== Now, find the one that hit most times "best"
alljbestV = jbestV(1);
hitjbestV = length(find(jbestV==jbestV(1)));
for k=2:tau_max
    if length(find(alljbestV==jbestV(k))) == 0
        alljbestV = [alljbestV; jbestV(k)];
        hitjbestV = [hitjbestV; length(find(jbestV==jbestV(k)))];
    end
end
[hitmax,jmax]=max(hitjbestV);
jbest = alljbestV(jmax);
%%%%% Generate surrogates with the "best" AR-model:
bV = bM(:,jbest); 
zM = zeros(n,M);
uM = zeros(n,M);
for k=1:M
    uV = zeros(n+p+ntrans,1);
    uV(1:p) = randn(p,1);
    for j=p+1:n+p+ntrans
        uV(j) = - bV(2:p+1)' * uV(j-1:-1:j-p) + randn;
    end
    uV = uV(p+ntrans+1:n+p+ntrans);
    [T,iuV] = sort(uV);
    [T,juV] = sort(iuV);
    zM(:,k) = oxV(juV); 
    uM(:,k) = uV; 
end
