function [zM,uM,bV] = stapanalytic(xV, m, p, M)
% [zM,uM,bV] = stapanalytic(xV, m, p, M)
% 
% STAPANALYTIC: Statically Transformed Autoregressive Process (STAP)
% using the analytic inverse gaussian transform. The rest is as in
% STAP function described below.
% STAP generates stochastic surrogates for a given time series as 
% statically transformed realisations of a Gaussian (autoregressive) 
% process. The generated surrogate data have the same amplitude 
% distribution (marginal cdf) and autocorrelation as the original time 
% series.
% INPUTS:
% -------
% xV	  : the given time series 
% m	      : order of the polynomial to approximate the sample transform
% p	      : order of the AR model to generate the Gaussian time series
% M       : number of surrogates to be generated
%
% OUTPUTS
% -------
% zM	  : matrix of generated x-like surrogate data ('M' columns)
% uM	  : matrix of generated Gaussian surrogate data ('M' columns)
% bV      : vector of parameters of the AR model
%
% Description of the algorithm and input parameters
% -------------------------------------------------
% First, the transform from marginal normal cdf to the marginal cdf of the 
% given time series 'xV' is estimated by a polynomial of order 'm' (the
% solution is the parameter vector 'aV'). Then the autocorrelation 'r_uV' 
% of the underlying AR-process is given as a function (defined as 
% "inverse transform") of the autocorrelation of 'xV', 'r_xV' (for lags
% up to 'p'). 
% This function is a polynomial also of order 'm' and the coefficients 'cV'
% are given analytically in terms of 'r_xV' and 'aV'. The parameters 'bV' 
% of the corresponding AR model of order 'p' are computed from 'r_uV' 
% using the Levinson algorithm. 
% A number of 'M' AR-time series 'uV' are generated and transformed to 'zV' 
% (by rank ordering), so that 'zV' has the same cdf as 'xV' and about the 
% same autocorrelation (however, 'r_zV' may vary substantially from 
% 'r_xV' at this stage).  

ntrans = 100;
n = length(xV);
[oxV,ixV] = sort(xV);
r_xV = xcorr(xV-mean(xV),p,'coeff');
r_xV(1:p) = [];  % the autocorrelation of xV
FxV = ([1:n]'-0.326)/(n+0.348); % The position plotting transform
owV = norminv(FxV);

%%% Unconstrained polynomial approximation (OLS) of the sample transform
pwxV = polyfit(owV,oxV,m);
pwxV = pwxV(:);
[invr_xM,cV] = invtranscorcoef(r_xV,flipud(pwxV));
%%% The polynomial transform for autocorrelations
invr_xV = zeros(p+1,1);
invr_xV(1) = r_xV(1);
for i=2:p+1
    inowV = find(imag(invr_xM(i,:))==0 & abs(invr_xM(i,:))<=1);
    if length(inowV) == 1
        successflag=1;
        invr_xV(i) = invr_xM(i,inowV);
    else
        successflag=0;
        if length(inowV)>1
            disp(['r_x(',int2str(i-1),') has ',int2str(length(inowV)),' solutions'])
        else   
            disp(['r_x(',int2str(i-1),') has no solutions']) 
        end
        break;
    end
end
%%%%% If a proper solution for r_u is found compute AR-parameters
if successflag == 1
    [bV,E]=levinson(invr_xV,p);
    bV = bV(:);
    rootbV = roots(bV);
    if any(abs(rootbV) >= 1)
        bV = polystab(bV);
        bV = bV(:);
        disp(['Root(s) > 1, polystab'])
        if any(abs(roots(bV)) >= 1)
            successflag = 0;
            disp(['Root(s) > 1, no solution'])
        end
    end
end
%%%%% If a proper AR model is found generate the surrogate time series
if successflag == 1
    zM = zeros(n,M);
    uM = zeros(n,M);
    for k=1:M
        uV = zeros(n+p+ntrans,1);
        uV(1:p) = randn(p,1);
        for j=p+1:n+p+ntrans
            uV(j) = - bV(2:p+1)' * uV(j-1:-1:j-p) + randn;
        end
        uV = uV(p+ntrans+1:n+p+ntrans);
        [T,iuV] = sort(uV);
        [T,juV] = sort(iuV);
        zM(:,k) = oxV(juV); 
        uM(:,k) = uV; 
    end
else
    disp('Generation of surrogate data failed')
end
